import json
import logging
import os
from json import JSONDecodeError
from typing import Callable, List, Literal, Optional

import jwt
import requests
from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest
from requests import Response

from ..models import User
from ..services import OpaService
from ..typing import JwtData
from ..utils.exceptions import AuthException, bad_gateway, forbidden, unauthorized

logger = logging.getLogger(__name__)

APP_IDENTIFIER = settings.IDP_USER_APP["APP_IDENTIFIER"]

ALLOWED_PATHS = settings.IDP_USER_APP.get("ALLOWED_PATHS", [])

USE_OPA = settings.IDP_USER_APP.get('USE_OPA', True)


class OpaAuthMiddleware:
    def __init__(self, get_response: Callable):
        self._get_response = get_response

    def __call__(self, request: WSGIRequest):

        try:
            # Require auth headers
            self._auth_required(request)
            # Get jwt payload from access token
            jwt_data = self._get_jwt_payload(request)
            # Check if required user claims are provided
            self._verify_jwt_claims(jwt_data, ['user_id', 'username'])
            # Check if user id added by ingress in the request is the same as the user id in the token
            self._verify_user_id(request, jwt_data)
            # Check if user is authorized to retrieve the resource
            self._authorize(request)

            # Set the user in the request for later access
            request.cardo_user = self._get_user(jwt_data['user_id'])

        except AuthException as exception:
            if request.path not in ALLOWED_PATHS:
                return exception.as_response()

        return self._get_response(request)

    @classmethod
    def _auth_required(cls, request: WSGIRequest):
        cls._access_token_required(request)
        cls._ingress_headers_required(request)

    @classmethod
    def _access_token_required(cls, request: WSGIRequest):
        cls._header_required(request, 'Authorization', 'token.', lambda x: x.startswith('Bearer '))

    @classmethod
    def _ingress_headers_required(cls, request: WSGIRequest):
        cls._header_required(request, 'X-USER-ID', 'user id.', lambda x: x.isdigit())
        cls._header_required(request, 'X-ROLES-FUNCTIONALITIES', 'roles functionalities.')

    @classmethod
    def _header_required(
            cls,
            request: WSGIRequest,
            header: str,
            header_human_name: str,
            validation_func: Optional[Callable] = None,
    ):
        header = cls._get_request_header(request, header)
        if not header:
            raise AuthException(unauthorized(f'Missing {header_human_name}'))
        if validation_func and not validation_func(header):
            raise AuthException(unauthorized(f'Invalid {header_human_name}'))

    @classmethod
    def _verify_user_id(cls, request: WSGIRequest, jwt_data: JwtData):
        if not int(cls._get_request_header(request, 'X-USER-ID')) == jwt_data['user_id']:
            raise AuthException(unauthorized('Invalid token.'))

    @classmethod
    def _get_access_token(cls, request: WSGIRequest) -> str:
        auth_header = cls._get_request_header(request, 'Authorization')
        return auth_header.replace('Bearer ', '')

    @classmethod
    def _get_jwt_payload(cls, request: WSGIRequest) -> JwtData:
        try:
            # Signature already verified from ingress before
            data = jwt.decode(cls._get_access_token(request), algorithms=['HS256'], options={'verify_signature': False})
        except Exception:
            raise AuthException(unauthorized('Invalid token.'))
        return data

    @classmethod
    def _verify_jwt_claims(cls, jwt_data: JwtData, claims: List[Literal['user_id', 'username']]):
        for claim in claims:
            cls._verify_jwt_claim(jwt_data, claim)

    @classmethod
    def _verify_jwt_claim(cls, jwt_data: JwtData, claim: Literal['user_id', 'username']):
        try:
            jwt_data[claim]
        except KeyError:
            raise AuthException(unauthorized('Invalid token.'))

    @classmethod
    def _authorize(cls, request: WSGIRequest):
        if not USE_OPA:
            return True

        response = cls._get_opa_response(request)
        if not response.ok:
            raise AuthException(bad_gateway('Cannot authorize at the moment.'))
        try:
            response_body = response.json()
        except JSONDecodeError:
            raise AuthException(bad_gateway('Cannot authorize at the moment.'))

        try:
            allow = response_body['result']
            if not allow:
                raise AuthException(forbidden())
            return True
        except KeyError:
            # If the key result is not present in the response,
            # it means the policy for this app does not exist in OPA
            # Update OPA with the necessary policy and data in this case
            OpaService.update_opa(authorization_header=request.headers.get('Authorization'))
            # Reattempt authorize
            return cls._authorize(request)

    @classmethod
    def _get_opa_response(cls, request: WSGIRequest) -> Response:
        opa_domain = settings.IDP_USER_APP['OPA_DOMAIN']
        opa_version = settings.IDP_USER_APP['OPA_VERSION']
        url = f"http://{opa_domain}/{opa_version}/data/{APP_IDENTIFIER}/{settings.APP_ENV}/allow"

        role = request.GET.get('role')
        roles_functionalities = json.loads(cls._get_request_header(request, 'X-ROLES-FUNCTIONALITIES'))
        allowed_functionalities = roles_functionalities.get(role, [])

        request_body = {
            'input': {
                'allowed_functionalities': allowed_functionalities,
                'path': cls._get_resource_path(request),
                'method': request.method,
                'role': role
            }
        }

        # Ask OPA for a policy decision
        return requests.post(url, json=request_body)

    @classmethod
    def _get_resource_path(cls, request: WSGIRequest) -> str:
        # Get the path as a list (removing leading and trailing /)
        request_path_as_list = request.path.strip('/').split('/')
        # Remove id values from path and add <id> as placeholder instead
        request_path_as_list = [p if not p.isdigit() else '<id>' for p in request_path_as_list]
        return '/'.join(request_path_as_list)

    @staticmethod
    def _get_request_header(request: WSGIRequest, header: str) -> str:
        return request.headers.get(header)

    @staticmethod
    def _get_user(user_id: int) -> User:
        try:
            return User.objects.get(idp_user_id=user_id)
        except User.DoesNotExist:
            raise AuthException(unauthorized())


class OpaAuthMiddlewareDev(OpaAuthMiddleware):

    def __call__(self, request: WSGIRequest):
        try:
            # Check if ingress headers are provided
            self._ingress_headers_required(request)

        except AuthException:
            if settings.IDP_USER_APP.get('USE_LOCAL_IDP_IN_DEV', False):
                # Ingress headers not available, call the IDP validate directly
                request = OpaAuthMiddlewareDev.__inject_headers_through_idp(request)
                return super().__call__(request)
            else:
                return self.__skip_auth_headers_and_opa(request)

    @staticmethod
    def __inject_headers_through_idp(request: WSGIRequest):
        response = requests.get(
            url=f"http://{os.getenv('IDP_URL')}/api/validate/?app={APP_IDENTIFIER}",
            headers={
                "Authorization": request.headers.get('Authorization'),
            }
        )

        # Headers dict cannot be modified, insert the headers as META,
        # that will be used by the overriden method _get_request_header
        request.META['X-USER-ID'] = response.headers.get('X-USER-ID')
        request.META['X-ROLES-FUNCTIONALITIES'] = response.headers.get('X-ROLES-FUNCTIONALITIES')

        return request

    @staticmethod
    def _get_request_header(request: WSGIRequest, header: str) -> str:
        return request.headers.get(header) or request.META.get(header)

    def __skip_auth_headers_and_opa(self, request: WSGIRequest):
        try:
            # Require access token
            self._access_token_required(request)
            # Get jwt payload from access token
            jwt_data = self._get_jwt_payload(request)
            # Check if required user claims are provided
            self._verify_jwt_claims(jwt_data, ['user_id', 'username'])
            # Set the user in the request for later access
            request.cardo_user = self._get_user(jwt_data['user_id'])

        except AuthException as exception:
            if request.path not in ALLOWED_PATHS:
                return exception.as_response()

        return self._get_response(request)



import faust
from asgiref.sync import sync_to_async
from django.conf import settings
from django.utils.module_loading import import_string
from faust import StreamT

from .services import UserService

app = import_string(settings.IDP_USER_APP['FAUST_APP_PATH'])

"""
{
    "idp_user_id": 12,
    "username": "",
    "first_name": "",
    "last_name": "",
    "email": "",
    "app_specific_configs": {
        "app_identifier": {
            "Servicer": {
                "app_config": {"vehicle_ids": [1, 2]},
                "permission_restrictions": {
                    "synchronizeDoD": False
                }
            }
        }
    }
}
"""


class UserRecord(faust.Record):
    idp_user_id: int
    first_name: str = None
    last_name: str = None
    username: str = None
    email: str = None
    app_specific_configs: dict = None


user_updates = app.topic(settings.IDP_USER_APP['USER_UPDATES_TOPIC_NAME'], value_type=UserRecord)


@sync_to_async
def update_user(user_record: UserRecord):
    UserService.update_user(user_record.asdict())


@app.agent(user_updates)
async def update_user_stream_processor(user_records: StreamT[UserRecord]):
    async for user_record in user_records:
        if user_record.app_specific_configs.get(settings.IDP_USER_APP['APP_IDENTIFIER']):
            await update_user(user_record)



from django.db import models


class User(models.Model):
    idp_user_id = models.BigIntegerField(primary_key=True)
    email = models.CharField(max_length=100)
    username = models.CharField(max_length=100, unique=True)
    first_name = models.CharField(max_length=100)
    last_name = models.CharField(max_length=100)



from rest_framework.views import exception_handler
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response, Response as DRFResponse
from rest_framework.status import HTTP_400_BAD_REQUEST, HTTP_401_UNAUTHORIZED, HTTP_403_FORBIDDEN, HTTP_502_BAD_GATEWAY


def bad_request(message="Bad request."):
    return Response({"message": message}, HTTP_400_BAD_REQUEST)


def unauthorized(message="Unauthorized."):
    return Response({"message": message}, HTTP_401_UNAUTHORIZED)


def forbidden(message="Forbidden."):
    return Response({"message": message}, HTTP_403_FORBIDDEN)


def bad_gateway(message="Bad Gateway."):
    return Response({"message": message}, HTTP_502_BAD_GATEWAY)


class AuthException(Exception):
    def __init__(self, response: DRFResponse):
        self._response = response
        self._response.accepted_renderer = JSONRenderer()
        self._response.accepted_media_type = 'application/json'
        self._response.content_type = 'application/json'
        self._response.renderer_context = {}

    def as_response(self) -> DRFResponse:
        return self._response.render()


def custom_exception_handler(exc, context):
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = exception_handler(exc, context)

    # Now add the HTTP status code to the response.
    if isinstance(exc, AuthException):
        return exc.as_response()

    return response



import pkgutil

import requests
from django.conf import settings
from rest_framework import status


class OpaService:
    @staticmethod
    def upload_policy_to_opa():
        policy_file = pkgutil.get_data(__name__, 'policy.rego')

        opa_domain = settings.IDP_USER_APP['OPA_DOMAIN']
        opa_version = settings.IDP_USER_APP['OPA_VERSION']
        app_identifier = settings.IDP_USER_APP['APP_IDENTIFIER']
        url = f"http://{opa_domain}/{opa_version}/policies/{app_identifier}/{settings.APP_ENV}"

        print(f"Uploading policy to {url}...")

        # Replace APP_IDENTIFIER and APP_ENV in rego file to match the app environment
        payload = policy_file.decode("utf-8") \
            .replace("{{APP_IDENTIFIER}}", app_identifier) \
            .replace("{{APP_ENV}}", settings.APP_ENV)
        response = requests.put(url, data=payload)

        if response.ok:
            print("Successfully uploaded policy")
        else:
            raise Exception(f"Could not upload policy!")

    @staticmethod
    def update_opa_data_through_idp(authorization_header):
        response = requests.post(
            url=f"http://{settings.IDP_USER_APP['IDP_URL']}/api/apps/update_opa_data/",
            json={'app': settings.IDP_USER_APP['APP_IDENTIFIER']},
            headers={
                "Authorization": authorization_header,
            }
        )

        if response.status_code != status.HTTP_204_NO_CONTENT:
            raise Exception("Cannot update OPA data through IDP!")

    @staticmethod
    def update_opa(authorization_header):
        OpaService.upload_policy_to_opa()
        OpaService.update_opa_data_through_idp(authorization_header)



# Generated by Django 3.1.5 on 2021-10-15 11:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('idp_user_id', models.BigIntegerField(primary_key=True, serialize=False)),
                ('email', models.CharField(max_length=100)),
                ('username', models.CharField(max_length=100, unique=True)),
                ('first_name', models.CharField(max_length=100)),
                ('last_name', models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='UserRole',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('role', models.CharField(max_length=125)),
                ('app_config', models.JSONField(null=True)),
                ('permission_restrictions', models.JSONField(default=dict)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_roles', to='idp_user.user')),
            ],
            options={
                'unique_together': {('user', 'role')},
            },
        ),
    ]



from typing import TypedDict, Union, List, Any


class JwtData(TypedDict):
    iat: int
    nbf: int
    jti: str
    exp: str
    type: str
    fresh: str
    user_id: int
    email: str
    username: str


class UserFeaturesPermissions(TypedDict):
    dod_manager: Union[List, bool]
    cash_flow_projection: Union[List, bool]
    notes_manager: Union[List, bool]


class AppSpecificConfigs(TypedDict):
    app_config: Any
    permission_restrictions: dict[str, Union[bool, Any]]


UserAppSpecificConfigs = dict[str, AppSpecificConfigs]


class UserUpdateEvent(TypedDict):
    idp_user_id: int
    first_name: str
    last_name: str
    username: str
    email: str
    app_specific_configs: UserAppSpecificConfigs


"""
"data": [
   {
        "idp_user_id": 12,
        "first_name": "str",
        "last_name": "str",
        "username": "str",
        "email": "str",
        "app_specific_configs": {
            "app_identifier": {
                "Servicer": {
                    "app_config": {"vehicle_ids": [1, 2]},
                    "permission_restrictions": {
                        "viewDoD": {"vehicle_ids": [1]},
                        "synchronizeDoD": false
                    }
                }
            }
        }
    }
]
"""
